/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdlib.h>  /* malloc(), free() */
#include <stdio.h>   /* printf() */
#include <assert.h>

#include <srf.h>
#include <builtins.h>
#include <function.h>

#ident "$Id: function.c,v 1.8 2004/11/02 04:39:39 bediger Exp $"

static struct function *free_list = NULL;

struct function *
new_function(void)
{
	struct function *r = NULL;
	if (free_list)
	{
		r = free_list;
		assert(NULL != r);
		free_list = r->next;
	} else {
		r = malloc(sizeof(*r));
		assert(NULL != r);;
	}

	r->function = NULL;
	r->debugflag = 0;
	r->type     = NOFUNC;
	r->identity_data.count = 0;
	r->identity_data.index = 0;
	r->composition_data.first    = NULL;
	r->composition_data.rest     = NULL;
	r->recursion_data.zero       = NULL;
	r->recursion_data.increment  = NULL;
	r->minimization_data.function= NULL;
	r->next     = NULL;

	return r;
}

void
free_function(struct function *f)
{
	if (f)
	{
		f->function = NULL;

		f->identity_data.count = 0;
		f->identity_data.index = 0;

		if (f->composition_data.first)
		{
			free_function(f->composition_data.first);
			f->composition_data.first = NULL;
		}

		if (f->composition_data.rest)
		{
			int i;
			for (i = 0; f->composition_data.rest[i]; ++i)
				free_function(f->composition_data.rest[i]);
			free(f->composition_data.rest);
			f->composition_data.rest = NULL;
		}

		if (f->recursion_data.zero)
		{
			free_function(f->recursion_data.zero);
			f->recursion_data.zero = NULL;
		}

		if (f->recursion_data.increment)
		{
			free_function(f->recursion_data.increment);
			f->recursion_data.increment = NULL;
		}

		if (f->minimization_data.function)
		{
			free_function(f->minimization_data.function);
			f->minimization_data.function = NULL;
		}

		f->next = free_list;
		free_list = f;
	}
}

struct function *
copy_function(struct function *original)
{
	struct function *f = NULL;

	if (original)
	{
		int i, j;

		f = new_function();

		f->function = original->function;

		f->debugflag = original->debugflag;

		f->type = original->type;

		f->identity_data.index = original->identity_data.index;
		f->identity_data.count = original->identity_data.count;

		f->composition_data.first
			= copy_function(original->composition_data.first);
		if (original->composition_data.rest)
		{
			for (i = 0; original->composition_data.rest[i]; ++i)
				;
			f->composition_data.rest = malloc(sizeof(struct function *) * (i + 1));
			for (j = 0; j < i; ++j)
			{
				f->composition_data.rest[j]
					= copy_function(original->composition_data.rest[j]);
			}
			f->composition_data.rest[j] = NULL;
		}

		f->recursion_data.zero
			= copy_function(original->recursion_data.zero);
		f->recursion_data.increment
			= copy_function(original->recursion_data.increment);

		f->minimization_data.function
			= copy_function(original->minimization_data.function);

		f->next = NULL;
	}

	return f;
}

void
print_function(struct function *f)
{
	if (f)
	{
		struct function **a;
		switch (f->type)
		{
		case NOFUNC:
			printf("No function");
			break;
		case SUCCESSOR:
			printf("s");
			break;
		case ZERO:
			printf("z");
			break;
		case IDENTITY:
			printf("id[%d,%d]",
				f->identity_data.count, f->identity_data.index+1);
			break;
		case COMPOSITION:
			printf("Cn[");
			print_function(f->composition_data.first);
			a = f->composition_data.rest;
			while (*a)
			{
				printf(",");
				print_function(*a);
				++a;
			}
			printf("]");
			break;
		case PREC:
			printf("Pr[");
			print_function(f->recursion_data.zero);
			printf(",");
			print_function(f->recursion_data.increment);
			printf("]");
			break;
		case MIN:
			printf("Mn[");
			print_function(f->minimization_data.function);
			printf("]");
			break;
		case CONSTRUCTOR:
			printf("Constructor");
			break;
		default:
			printf("Bad builtinType: %d\n", f->type);
			break;
		}
	}
}

void
free_functions(void)
{
	while (free_list)
	{
		/* fix up free_list here, since subsequent code calls
		 * free_function() which pushes structs function onto
		 * free_list */
		struct function *tmp = free_list;
		free_list = free_list->next;
		tmp->next = NULL;

		if (tmp->composition_data.first)
		{
			free_function(tmp->composition_data.first);
			tmp->composition_data.first = NULL;
		}
		if (tmp->composition_data.rest)
		{
			unsigned int i;
			for (i = 0; tmp->composition_data.rest[i]; ++i)
			{
				free_function(tmp->composition_data.rest[i]);
				tmp->composition_data.rest[i] = NULL;
			}
			free(tmp->composition_data.rest);
			tmp->composition_data.rest = NULL;
		}
		if (tmp->recursion_data.zero)
		{
			free_function(tmp->recursion_data.zero);
			tmp->recursion_data.zero = NULL;
		}
		if (tmp->recursion_data.increment)
		{
			free_function(tmp->recursion_data.increment);
			tmp->recursion_data.increment = NULL;
		}
		free(tmp);
	}
}

