%{
/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdio.h>  /* fprintf(), printf() */
#include <stdlib.h> /* malloc(), free() */
#include <unistd.h> /* getopt() */
#include <string.h> /* strerror() */
#include <errno.h>  /* errno declaration */
#include <assert.h> /* assert macro */
#include <signal.h> /* assert macro */
#include <setjmp.h> /* setjmp(), longjmp(), jmp_buf */
#include <sys/time.h> /* gettimeofday(), struct timeval */

#ident "$Id: grammar.y,v 1.18 2004/11/02 04:39:39 bediger Exp $"

#include <srf.h>
#include <hashtable.h>
#include <value.h>
#include <function.h>
#include <builtins.h>
#include <atom.h>

int yyerror(char *s1);

void handle_value(struct value *topmost_value);
struct value *handle_function(struct function *f, struct value_list *args);
struct value *set_debug(struct value *val);
struct value *set_timeout(struct value *val);
void sigint_handler(int signo);

jmp_buf in_function_eval;
int eval_timeout = 0;

extern int primitive_call_counter;
int function_timer = 0;
int function_counter = 0;
void set_flag(const char *thinger);
float elapsed_time(struct timeval before, struct timeval after);

/* Imports from lex.l */
extern void push_and_open(const char *filename);
extern void set_yyin(const char *filename);
extern void reset_yyin(void);
extern void set_yyin_stdin(void);

%}

%union {
	const char *identifier;
	const char *string_constant;
	struct value *value;
	struct value_list *value_list;
}

%token TK_LPAREN TK_RPAREN 
%token TK_LBRACKET TK_RBRACKET 
%token <identifier> TK_IDENTIFIER
%token <identifier> TK_CONSTRUCTOR
%token TK_EQUALS
%token TK_SEMI TK_COMMA
%token TK_LOAD TK_DEBUG TK_TIMEOUT
%token <string_constant> TK_MODIFIER
%token <string_constant> STRING_CONSTANT

%type <value> program error
%type <value> stmnt value function_call constructor
%type <value_list> value_list

%%


program
	: stmnt TK_SEMI { handle_value($1); }
	| program stmnt TK_SEMI { handle_value($2); }
	| error  /* magic token that causes yacc to unwind to on syntax error */
	;

stmnt
	: TK_IDENTIFIER TK_EQUALS value
		{
			struct value *p, *in = $3;
			if (IDENTIFIER == in->type)
			{
				struct value *tmp = value_lookup(in->v.identifier);
				if (tmp)
				{
					free_value(in);
					in = tmp;
				}
			}

			if (in->assigned)
				in = copy_value(in);

			in->assigned = 1;

			p = (struct value *)add_data(value_table, $1, in);
			if (p)
			{
				p->assigned = 0;
				free_value(p);
			}
			$$ = in;

			/* Don't ever want to call free_value() on $3
			 * Either it has assigned == 1, or we mark it that way. */
		}
	| TK_IDENTIFIER { $$ = value_lookup($1); }
	| function_call { $$ = $1; }
	| constructor   { $$ = $1; }
	| TK_LOAD STRING_CONSTANT
		{
			$$ = new_value();
			$$->type = FILENAME;
			$$->v.filename = $2;
		}
	| TK_DEBUG value { $$ = set_debug($2); }
	| TK_TIMEOUT value { $$ = set_timeout($2); }
	| { $$ = NULL; }  /* empty production to allow for lonely ';' */
	;

function_call
	: TK_IDENTIFIER TK_LPAREN value_list TK_RPAREN
		{
			struct value *v = value_lookup($1);
			if (v)
			{
				if (FUNCTION == v->type)
				{
					$$ = handle_function(v->v.function, $3);
				} else {
					printf("\"%s\": not a function\n", $1);
					$$ = NULL;
					free_value_list($3);
				}
			} else {
				free_value_list($3);
				printf("%s: no such function\n", $1);
				$$ = NULL;
			}
		}
	| TK_MODIFIER TK_IDENTIFIER TK_LPAREN value_list TK_RPAREN
		{
			struct value *v = value_lookup($2);
			if (v)
			{
				if (FUNCTION == v->type)
				{
					set_flag($1);
					$$ = handle_function(v->v.function, $4);
				} else {
					printf("\"%s\": not a function\n", $2);
					$$ = NULL;
					free_value_list($4);
				}
			} else {
				free_value_list($4);
				printf("%s: no such function\n", $2);
				$$ = NULL;
			}
		}
	| constructor TK_LPAREN value_list TK_RPAREN
		{
			$$ = handle_function($1->v.function, $3);
			free_value($1);
		}
	| TK_CONSTRUCTOR TK_LPAREN value_list TK_RPAREN
		{
			/* production to catch common errors - not really part
			 * of the grammar. */
			$$ = new_value();
			free_value_list($3);
			fprintf(stderr, "%s: not a function\n", $1);
		}
	| TK_IDENTIFIER TK_LBRACKET value_list TK_RBRACKET
		{
			/* production to catch common errors - not really part
			 * of the grammar. */
			$$ = new_value();
			free_value_list($3);
			fprintf(stderr, "%s: not a constructor\n", $1);
		}
	;

constructor
	: TK_CONSTRUCTOR TK_LBRACKET value_list TK_RBRACKET
		{
			struct value *v = value_lookup($1);
			if (v)
			{
				if (FUNCTION == v->type)
				{
					$$ = handle_function(v->v.function, $3);
				} else {
					printf("\"%s\": not a function\n", $1);
					$$ = NULL;
					free_value_list($3);
				}
			} else {
				free_value_list($3);
				printf("%s: no such function constructor\n", $1);
				$$ = NULL;
			}
		}
	;

value_list
	: value
		{
			$$ = new_value_list();
			$$->tail = $1;
			$$->head = $1;
			$$->cnt  = 1;
		}
	| value_list TK_COMMA value
		{
			$1->tail->next = $3;
			$1->tail = $1->tail->next;
			++$1->cnt;
			$$ = $1;
		}

	|
		{
			/* does this need to happen? */
			$$ = new_value_list();
		}

	;

value
	: TK_IDENTIFIER
		{
			$$ = new_value();
			$$->v.identifier = $1;
			$$->type = IDENTIFIER;
		}

	| function_call { $$ = $1; }
	| constructor   { $$ = $1; }
	;
%%

struct filename_node {
	const char *filename;
	struct filename_node *next;
};

void
usage(char *progname)
{
	fprintf(stderr,
		"%s: a simple recursive function interpreter\n"
		"Usage: %s [-L filename [-L otherfile ...]] [-t timeout]\n"
		"",
		progname,
		progname
	);
}


int
main(int ac, char **av)
{
	int r, flag;
	struct hashtable *h;
	struct filename_node *p;
	struct filename_node *load_files = NULL, *load_tail = NULL;
	extern const char *current_input_stream;
	extern char *optarg;
	
	h = init_hashtable(256, 10);
	setup_atom_table(h);
	setup_value_table(h);
	setup_builtins();
	current_input_stream = Atom_string("stdin");

	while (-1 != (flag = getopt(ac, av, "L:t:xXh")))
	{
		switch (flag)
		{
		case 'L':
			p = malloc(sizeof(*p));
			p->filename = Atom_string(optarg);
			p->next = NULL;
			if (load_tail)
				load_tail->next = p;
			load_tail = p;
			if (!load_files)
				load_files = p;
			break;
		case 't':
			eval_timeout = atoi(optarg);
			break;
		case 'x':
		case 'X':
		case 'h':
			usage(av[0]);
			exit(0);
			break;
		}
	}

	if (load_files)
	{
		struct filename_node *t, *z;
		for (z = load_files; z; z = t)
		{
			FILE *fin;

			t = z->next;

			printf("load file named \"%s\"\n",
				z->filename);

			if (!(fin = fopen(z->filename, "r")))
			{
				fprintf(stderr, "Problem reading \"%s\": %s\n",
					z->filename, strerror(errno));
				continue;
			}

			set_yyin(z->filename);

			r = yyparse();

			reset_yyin();

			if (r)
				printf("Problem with file \"%s\"\n", z->filename);

			free(z);
			fin = NULL;
		}
	}

	set_yyin_stdin();

	do {
		r = yyparse();
	} while (r);

	free_hashtable(h);
	free_values();
	free_functions();
	return r;
}

void
handle_value(struct value *v)
{
	if (NULL != v)
	{
		if (FILENAME == v->type)
			push_and_open(v->v.filename);
		else 
			print_value(v);
		free_value(v);
	}
}

int
yyerror(char *s1)
{
	extern int lineno;
	extern const char *current_input_stream;

	fprintf(stderr, "%s, line %d, %s\n", current_input_stream, lineno, s1);

	return 0;
}

struct value *
set_timeout(struct value *val)
{
	struct value *r = val;
	int t = 0;

	if (val)
	{
		if (IDENTIFIER == val->type)
		{
			struct value *tmp = value_lookup(val->v.identifier);
			val = tmp;
			/* don't free_value(val) here - it might get passed
			 * back to caller */
		}

		if (INTEGER == val->type)
			t = val->v.value;
	}

	eval_timeout = t;

	return r;
}

struct value *
set_debug(struct value *val)
{
	struct value *r = val;

	if (val)
	{
		if (IDENTIFIER == val->type)
		{
			struct value *tmp = value_lookup(val->v.identifier);
			val = tmp;
			/* don't free_value(val) here - it might get passed
			 * back to caller */
		}

		if (FUNCTION == val->type)
			val->v.function->debugflag = 1;
	}

	return r;
}

/* This function only called at "top level" of interpreter.
 * It just "wraps" evaluate_function() - it sets a signal
 * handler and jmp_buf, and does a gettimeofday() before calling
 * evaluate_function(), then resets the signal handler after
 * evaluate_function() returns.  If specific flag variables
 * evaluate to "true", it prints elapsed time or primitive call count.
 * Can't incorporate this into evaluate_function()
 * because the jmp_buf should only get set to "come back"
 * to this level of the C call stack. */
struct value *
handle_function(struct function *f, struct value_list *args)
{
	struct value *r = NULL;
	int cc;
	void (*old_sigint_handler)(int);
	void (*old_sigalm_handler)(int);
	struct timeval before, after;

	old_sigint_handler = signal(SIGINT, sigint_handler);
	old_sigalm_handler = signal(SIGALRM, sigint_handler);

	primitive_call_counter = 0;

	if (!(cc = setjmp(in_function_eval)))
	{
		alarm(eval_timeout);
		gettimeofday(&before, NULL);
		r = evaluate_function(f, args);
		alarm(0);
		gettimeofday(&after, NULL);
	} else {
		alarm(0);
		gettimeofday(&after, NULL);
		printf("%s\n", cc == 1? "Interrupt": "Timeout");
		r = new_value();
	}

	signal(SIGINT, old_sigint_handler);
	signal(SIGALRM, old_sigalm_handler);

	if (function_timer)
	{
		printf("elapsed time %.3f\n", elapsed_time(before, after));
		function_timer = 0;
	}

	if (function_counter)
	{
		printf("primitive count: %d\n", primitive_call_counter);
		function_counter = 0;
	}

	return r;
}

void
sigint_handler(int signo)
{
	longjmp(in_function_eval, signo == SIGINT? 1: 2);
}

/* Set global flag variables function_timer and function_counter
 * appropriately. */
void
set_flag(const char *thinger)
{
	static const char *time_atom = NULL;
	static const char *count_atom = NULL;

	if (!time_atom)
		time_atom = Atom_string("time");
	if (!count_atom)
		count_atom = Atom_string("count");

	/* Hey, these might constitute the only times in this program
	 * that an "atom" comparison happens. */
	if (thinger == time_atom)
		function_timer = 1;

	if (thinger == count_atom)
		function_counter = 1;
}

/* utility function elapsed_time() */
float
elapsed_time(struct timeval before, struct timeval after)
{
	float r = 0.0;

	if (before.tv_usec > after.tv_usec)
	{
		after.tv_usec += 1000000;
		--after.tv_sec;
	}

	r = (float)(after.tv_sec - before.tv_sec)
		+ (1.0E-6)*(float)(after.tv_usec - before.tv_usec);

	return r;
}
