
%{
/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdlib.h>
#include <string.h>   /* strdup() */
#include <errno.h>    /* errno declaration */

#include <hashtable.h>
#include <atom.h>
#include <srf.h>

#include "srf-grammar.h"

static char *rcsident = "$Id: lex.l,v 1.14 2004/11/02 04:39:39 bediger Exp $";

int lineno = 1;

int reading_file = 0;

const char *current_input_stream;

char *unescape_string(char *s);
void  push_and_open(const char *filename);

struct stream_node {
#ifdef FLEX_SCANNER
	YY_BUFFER_STATE stream;
#else
	FILE *stream;
#endif
	struct stream_node *next;
	const char *old_filename;
	int old_lineno;
};

/* XXX - what to do with that at exit time? */
struct stream_node *file_stack = NULL;

%}

%%

#..*$   { ++lineno; /* don't echo comments */ }
"file"  { reading_file = !reading_file; }
"debug"	{ return TK_DEBUG; }
"load"	{ return TK_LOAD; }
"timeout"	{ return TK_TIMEOUT; }
\"(\\.|[^\\"])*\" {
	char *tmp;
	if (reading_file) ECHO;
	tmp = unescape_string(yytext);
	yylval.string_constant = Atom_string(tmp);
	free(tmp);
	return(STRING_CONSTANT);
}
[ \t]	{ /* ignore whitespace */ if (reading_file) ECHO; }
[\n]	{ ++lineno; if (reading_file) ECHO; }
Mn|Cn|id|Pr {
	/* these identify "functors", things to build up complex functions */
	if (reading_file) ECHO;
	yylval.identifier = Atom_string(yytext);
	return TK_CONSTRUCTOR;
};
time|count {
	yylval.string_constant = Atom_string(yytext);
	return TK_MODIFIER;
}
[^ \t\n()\[\];,=#"][^ \t\n()\[\];,=#"]* {
	/* Extremely generous "identifier" regexp - basically only
	 * disallows characters like '(', ']', ';' that have special
	 * meaning elsewhere. */
	if (reading_file) ECHO;
	yylval.identifier = Atom_string(yytext);
	return TK_IDENTIFIER;
};
\(		{ if (reading_file) ECHO; return TK_LPAREN; }
\)		{ if (reading_file) ECHO; return TK_RPAREN; }
\[		{ if (reading_file) ECHO; return TK_LBRACKET; }
\]		{ if (reading_file) ECHO; return TK_RBRACKET; }
\=		{ if (reading_file) ECHO; return TK_EQUALS; }
\;		{ if (reading_file) ECHO; return TK_SEMI; }
\,		{ if (reading_file) ECHO; return TK_COMMA; }
.		{ if (reading_file) ECHO; /* Just eat it up */ }

%%
char *
unescape_string(char *s)
{
	char *n = NULL;
	if (s)
	{
		int i, j, l;
		if ('"' == s[0])
			++s;
		if ('"' == s[strlen(s) - 1])
			s[strlen(s) - 1] = '\0';
		n = malloc(strlen(s) + 1);
		l = strlen(s);
		for (i = 0, j = 0; i < l; ++i)
		{
			if ('\\' == s[i])
			{
				++i;
				switch (s[i])
				{
				/* XXX - doesn't do '\0' or other numerical escapes */
				case 't': n[j++] = '\t';  break;
				case 'r': n[j++] = '\r';  break;
				case 'n': n[j++] = '\n';  break;
				case '\\': n[j++] = '\\'; break;
				default:
					n[j++] = s[i];
					break;
				}
			} else {
				n[j++] = s[i];
			}
		}
		n[j] = '\0';
	}
	return n;
}

void
push_and_open(const char *filename)
{
	FILE *fin;
	extern FILE *yyin;
	extern int reading_file, lineno;
	extern const char *current_input_stream;

	if (NULL != (fin = fopen(filename, "r")))
	{
		struct stream_node *n;
		n = malloc(sizeof(*n));
#ifdef FLEX_SCANNER
		n->stream = YY_CURRENT_BUFFER;
		yyin = fin;
		yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
#else
		n->stream = yyin;
		yyin = fin;
#endif
		n->next = file_stack;
		n->old_filename = current_input_stream;
		n->old_lineno = lineno;
		current_input_stream = filename;
		file_stack = n;
		lineno = 0;
		/* ++reading_file; */
	} else {
		fprintf(stderr, "Could not open \"%s\" for read: %s\n",
			filename, strerror(errno));
	}
}

void
set_yyin_stdin(void)
{
		yyin = stdin;
#ifdef FLEX_SCANNER
		yy_delete_buffer(YY_CURRENT_BUFFER);
		yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
#endif
}

void
set_yyin(const char *filename)
{
	FILE *fin;

	if (NULL != (fin = fopen(filename, "r")))
	{
		yyin = fin;
#ifdef FLEX_SCANNER
		yy_delete_buffer(YY_CURRENT_BUFFER);
		yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
#endif
		current_input_stream = filename;
		lineno = 0;
		/* ++reading_file; */
	} else {
		fprintf(stderr, "Could not open \"%s\" for read: %s\n",
			filename, strerror(errno));
	}
}

void
reset_yyin(void)
{
	if (yyin)
		fclose(yyin);
	yyin = NULL;
#ifdef FLEX_SCANNER
	yy_delete_buffer(YY_CURRENT_BUFFER);
#endif
}

int
yywrap()
{
	extern FILE *yyin;
	extern int reading_file;
	extern int lineno;
	extern struct stream_node *file_stack;

	int r = 1;
	if (file_stack)
	{
		struct stream_node *tmp = file_stack->next;
		fclose(yyin);
#ifdef FLEX_SCANNER
		yy_delete_buffer(YY_CURRENT_BUFFER);
		yy_switch_to_buffer(file_stack->stream);
#else
		yyin = file_stack->stream;
#endif
		current_input_stream = file_stack->old_filename;
		lineno = file_stack->old_lineno;
		file_stack->stream = NULL;
		file_stack->next = NULL;
		free(file_stack);
		file_stack = tmp;
		r = 0;
		/* --reading_file; */
	} else {
		/* printf("No files on stack, returning 1\n"); */
	}
	return r;
}
