%{
/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdio.h>  /* fprintf(), printf() */
#include <stdlib.h> /* malloc() */
#include <unistd.h> /* getopt() */

#include <hashtable.h>

#ident "$Id: gram2.y,v 1.6 2004/11/02 04:39:45 bediger Exp $"

char outbuf[256];

extern char *optarg;

enum NodeType { UNSET, OPERATION, CONSTANT, VARIABLE, FUNCTION };

struct function {
	unsigned char *fname;
	int child_count;
	struct node *children;  /* list, not array */
};

struct node {
	enum NodeType ntype;
	union {
		unsigned char op;
		unsigned int  con;
		unsigned char *var;
		struct function *fun;
	} val;
	struct node *left;
	struct node *right;
	struct node *next;
};

void print_constant(unsigned int constant);
void print_operation(struct node *n);
void assign_indexes(struct node *n);
void walk_tree(struct node *n);
struct node *copy_node(struct node *existing_node);
struct node *new_variable_node(unsigned char *ident);
struct node *new_const_node(unsigned int con);
struct node *new_function_node(unsigned char *fname, struct node *args);
struct node *new_op_node(unsigned char op, struct node *left, struct node *right);
struct node *opposite_relation(struct node *relation_node);
void replace_names(struct node *n, struct hashtable *h);
void massage_names(struct hashtable *h);
void collect_names(struct node *n, struct hashtable *h);
void print_function(struct function *f);
int  cover(const void *a, const void *b);

%}

%union {
	unsigned int   con;
	unsigned char *ident;
	struct node   *node;
}

%token <ident> TK_IDENTIFIER
%token <con> TK_CONST
%token TK_LPAREN TK_RPAREN TK_COMMA
%token TK_PLUS TK_MINUS TK_TIMES TK_DIV
%token TK_SEMI TK_IF TK_ELSE
%token <ident> TK_LOGICAL_OP

%type <node> program stmnt additive_expr multiplicative_expr function_call
%type <node> value value_list relation

%%

/* these productions exist so that yacc will generate code
 * capable of reading more than a single stmnt - a human
 * can type in many things, refining an answer.
 */
program
	: stmnt TK_SEMI { assign_indexes($1); walk_tree($1); printf(";\n"); }
	| program stmnt TK_SEMI { assign_indexes($2); walk_tree($2); printf(";\n"); }
	;


stmnt
	: additive_expr { $$ = $1; }
	| TK_IF TK_LPAREN relation TK_RPAREN stmnt TK_ELSE stmnt
		{
			/* In terms of simple recursive functions, you have to
			 * have the "else" clause as well - otherwise you make a partial
			 * function, right? */

			/* this amounts to writing a function for:
			 *  relation * stmnt1 + opposite-of-relation * stmnt2
			 *  Cn[sum,
			 *  	Cn[prod, relation, stmnt1],
			 *  	Cn[prod, !relation, stmnt2]
			 * ];
			 *
			 * Note that this makes an if-then-else into a stmnt that
			 * the program can evaluate - if-then-else has a value, either
			 * the value of the true clause or the value of the false clause.
			 * Both clauses get evaluated as does "relation" and its opposite,
			 * due to the nature of recursive functions.
			 */

			struct node *x, *y, *z;

			x = new_op_node('*', $3, $5);
			y = opposite_relation($3);
			z = new_op_node('*', y, $7);
			$$ = new_op_node('+', x, z);
		}
	;

relation
	: additive_expr TK_LOGICAL_OP additive_expr
		{
fprintf(stderr, "relation, %s\n", $2);
			/* lexically limit the logical operations to a well-known
			 * set { <, >, <=, >=, ==, != } so that this program can easily
			 * come up with the "oppposite" later to make the "else" clause
			 * of an if-then-else construct work in simple recursive functions.
			 * Other than that, no reason exists to limit the function whose
			 * gets tested by "if".
			 */
			$1->next = NULL;
			$3->next = $1;
			$$ = new_function_node($2, $3);
		}
	;

additive_expr
	:                        multiplicative_expr  { $$ = $1; }
	| additive_expr TK_PLUS  multiplicative_expr  { $$ = new_op_node('+', $1, $3); }
	| additive_expr TK_MINUS multiplicative_expr  { $$ = new_op_node('-', $1, $3); }
	;

multiplicative_expr
	:                              value  { $$ = $1; }
	| multiplicative_expr TK_TIMES value  { $$ = new_op_node('*', $1, $3); }
	| multiplicative_expr TK_DIV   value  { $$ = new_op_node('/', $1, $3); }
	;

function_call
	: TK_IDENTIFIER TK_LPAREN value_list TK_RPAREN { $$ = new_function_node($1, $3); }
	;

value_list
	: additive_expr                      { $$ = $1; }
	| value_list TK_COMMA additive_expr  { $3->next = $1; $$ = $3; }
	;

value
	: TK_CONST       { $$ = new_const_node($1); }
	| TK_IDENTIFIER  { $$ = new_variable_node($1); }
	| function_call  { $$ = $1; }
	| TK_LPAREN additive_expr TK_RPAREN  { $$ = $2; }
	;

%%

int
yyerror(s1)
	char *s1;
{
	extern int lineno;

	fprintf(stderr, "line %d, %s\n", lineno, s1);

	return 0;
}

struct node *
new_node(void)
{
	struct node *n = malloc(sizeof(*n));
	n->ntype = UNSET;
	n->val.var = NULL;
	n->next = n->left = n->right = NULL;
	return n;
}

struct node *
new_op_node(unsigned char op, struct node *left, struct node *right)
{
	struct node *n = new_node();
	n->ntype = OPERATION;
	n->val.op = op;
	n->left = left;
	n->right = right;
	return n;
}

struct node *
new_function_node(unsigned char *fname, struct node *args)
{
	struct node *n = new_node();
	struct node *last;

	n->ntype = FUNCTION;
	n->val.fun = malloc(sizeof(struct function));
	n->val.fun->fname = fname;
	n->val.fun->child_count = 0;

	/* reverse the argument list - the grammar "pushes" arguments
	 * on the list, ending up with an incorrectly-ordered list */
	last = NULL;
	while (args)
	{
		struct node *tmp = args;
		args = args->next;
		++n->val.fun->child_count;
		if (tmp)
		{
			tmp->next = last;
			last = tmp;
		}
	}

	n->val.fun->children = last;

	return n;
}

struct node *
new_const_node(unsigned int con)
{
	struct node *n = new_node();
	n->ntype = CONSTANT;
	n->val.con = con;
	return n;
}

struct node *
new_variable_node(unsigned char *ident)
{
	struct node *n = new_node();
	n->ntype = VARIABLE;
	n->val.var = ident;
	return n;
}

struct node *
opposite_relation(struct node *relation_node)
{
	/* not much consistency checking - grammar and lexer enforce
	 * relation_node->ntype == FUNCTION, and that 
	 * relation_node->val.fun->fname evaluates to one of a very
	 * few values. */
	struct node *n;
	unsigned char *relation = relation_node->val.fun->fname;
	unsigned char *opposite = NULL;

	if (!strcmp("<", relation))
		opposite = ">=";
	else if (!strcmp(">", relation))
		opposite = "<=";
	else if (!strcmp(">=", relation))
		opposite = "<";
	else if (!strcmp("<=", relation))
		opposite = ">";
	else if (!strcmp("==", relation))
		opposite = "!=";
	else if (!strcmp("!=", relation))
		opposite = "==";

	n = copy_node(relation_node);

	n->val.fun->fname = opposite;

	return n;
}

void
walk_tree(struct node *n)
{
	switch (n->ntype)
	{
	case UNSET:
		break;
	case OPERATION:
		print_operation(n);
		break;
	case CONSTANT:
		print_constant(n->val.con);
		break;
	case VARIABLE:
		printf("%s", n->val.var);  /* gets replaced by id[x,y] functor */
		break;
	case FUNCTION:
		print_function(n->val.fun);
		break;
	}
}

struct node *
copy_node(struct node *old)
{
	struct node *n = NULL;
	if (old)
	{
		switch (old->ntype)
		{
		case UNSET:
			n = new_const_node(0);
			n->ntype = UNSET;
			break;
		case OPERATION:
			n = new_op_node(
					old->val.op,
					copy_node(old->left),
					copy_node(old->right)
				);
			break;
		case CONSTANT:
			n = new_const_node(old->val.con);
			break;
		case VARIABLE:
			n = new_variable_node(old->val.var);
			break;
		case FUNCTION:
			n = new_function_node(old->val.fun->fname, NULL);
			n->val.fun->child_count = old->val.fun->child_count;
			n->val.fun->children = copy_node(old->val.fun->children);
			break;
		}
		n->next = copy_node(old->next);
	}
	return n;
}

void
print_operation(struct node *n)
{
	const char *name = "unset";
	switch (n->val.op)
	{
	case '+': name = "sum";  break;
	case '-': name = "sub";  break;
	case '*': name = "prod"; break;
	case '/': name = "div";  break;
	/* XXX - need to get remainder (mod) in here */
	default:
	/* XXX - need to handle default case */
		break;
	}
	printf("Cn[ %s,", name);
	if (n->left) walk_tree(n->left);
	printf(", ");
	if (n->right) walk_tree(n->right);
	printf("]");
}

void
print_constant(unsigned int constant)
{
	if (constant)
	{
		printf("Cn[s, ");
		print_constant(constant - 1);
		printf("]");
	} else {
		printf("z");
	}
}

/* Walk a tree of structs node, collect all the variable names.
 * Walk the tree again, replacing variable names with identity
 * functors.
 * Assume that variable names that appear in the tree derive from
 * a lexically-ordered list of arguments in a function call.
 */
void
assign_indexes(struct node *n)
{
	struct hashtable *h;

	h = init_hashtable(256);

	collect_names(n, h);
	massage_names(h);
	replace_names(n, h);

	free_hashtable(h);
}

/* Collect all the variable names appearing in the expression to turn into
 * a function */
void
collect_names(struct node *n, struct hashtable *h)
{
	if (n->left) collect_names(n->left, h);
	if (FUNCTION == n->ntype)
	{
		struct node *k = n->val.fun->children;
		for (; k; k = k->next)
			collect_names(k, h);
	} else if (VARIABLE == n->ntype) {
		if (!find_node(h, (char *)n->val.var))
			insert_value(h, (char *)n->val.var, NULL);
	}
	if (n->right) collect_names(n->right, h);
}

void
replace_names(struct node *n, struct hashtable *h)
{
	if (n->left) replace_names(n->left, h);
	if (FUNCTION == n->ntype)
	{
		struct node *k = n->val.fun->children;
		for (; k; k = k->next)
			replace_names(k, h);
	} else if (VARIABLE == n->ntype) {
		unsigned char *new_name = find_value(h, (char *)n->val.var);
		n->val.var = new_name;
	}
	if (n->right) replace_names(n->right, h);
}

void
massage_names(struct hashtable *h)
{
	char **names;
	int arg_count;
	int i, j;

	arg_count = h->size;

	/* I predicate using an array to store the variable names 
	 * on the assumption that any expression won't have too
	 * many of them, so an elaborate data structure won't offer
	 * too much efficiency */
	names = malloc(arg_count * sizeof(*names));

	/* collect all the variable names in the hashtable by
	 * putting them into the array, "names". */
	for (i = 0, j = 0; i < h->count; ++i)
	{
		struct hashnode *chain;
		for (chain = h->buckets[i]->next; chain; chain = chain->next)
		{
			if (chain->key)
				names[j++] = chain->key;
		}
	}

	/* This is where the program assumes
	 * that the desired recursive function has lexically ordered
	 * arguments. */
	qsort(names, arg_count, sizeof(*names), cover);

	/* Fix up the hashtable - for all keys, put in a value that correspnds
	 * to an identity functor. replace_names() traverses the parse tree
	 * again, and replaces all variable names with the functors constructed
     * here. */
	for (i = 0; i < h->count; ++i)
	{
		struct hashnode *chain;
		for (chain = h->buckets[i]->next; chain; chain = chain->next)
		{
			if (chain->key)
			{
				for (j = 0; j < arg_count; ++j)
				{
					if (!strcmp(names[j], chain->key))
					{
						char buf[128];
						snprintf(buf, sizeof(buf),
							"id[%d,%d]", arg_count, j + 1
						);
						chain->value = malloc(strlen(buf) + 1);
						strcpy(chain->value, buf);
						break;
					}
				}
			}
		}
	}
}

char *
function_map(char *original)
{
	char *r = original;
	if (!strcmp(original, ">"))
		r = "greaterthan";
	else if (!strcmp(original, "<"))
		r = "lessthan";
	else if (!strcmp(original, "<="))
		r = "lessthan_or_equalto";
	else if (!strcmp(original, ">="))
		r = "greaterthan_or_equalto";
	else if (!strcmp(original, "=="))
		r = "equalto";
	else if (!strcmp(original, "!="))
		r = "notequalto";
	return r;
}

void print_function(struct function *f)
{
	struct node *t;
	printf("Cn[%s", function_map(f->fname));
	for (t = f->children; t; t = t->next)
	{
		printf(", ");
		walk_tree(t);
	}
	printf("]");
}

/* cover up strcmp() use in the qsort() call - the parameters
 * seem overly complicated, but it works. */
int
cover(const void *a, const void *b)
{
	char **x = (char **)a;
	char **y = (char **)b;
	return strcmp(*x, *y);
}
