/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ident "$Id: hashtable.c,v 1.4 2004/11/02 04:39:45 bediger Exp $"
#include <stdio.h>
#include <stdlib.h>  /* malloc(), free() */
#include <string.h>  /* strlen(), strcpy(), strcmp() */
#include <assert.h>  /* assert macro */

#include <hashtable.h>

unsigned long hash(unsigned char *str);

/* number of buckets has to be a power of 2 for this to work */
#define MOD(x,y) ((x)&((y)-1))

struct hashtable *
init_hashtable(int size)
{
	unsigned int i;
	struct hashtable *h = malloc(sizeof(*h));
	h->count = size;
	h->size  = 0;
	h->buckets = malloc(sizeof(struct hashnode *) * h->count);
	for (i = 0; i < h->count; ++i)
	{
		struct hashnode *head, *tail;

		head = malloc(sizeof(*head));
		tail = malloc(sizeof(*tail));

		head->next = tail;
		tail->prev = head;
		head->prev = NULL;
		tail->next = NULL;

		head->value = tail->value = NULL;
		head->key   = tail->key   = NULL;

		h->buckets[i] = head;
	}

	return h;
}

struct hashnode *
find_node(struct hashtable *h, char *key)
{
	struct hashnode *chain;

	if (h->size)
	{
		unsigned int index;
		unsigned long hv = hash((unsigned char *)key);


		index = MOD(hv, h->count);
		chain = h->buckets[index]->next;

		while (NULL != chain)
		{
			if (chain->key && !strcmp(chain->key, key))
				break;
			chain = chain->next;
		}
	} else
		chain = NULL;
	
	return chain;
}

void *
find_value(struct hashtable *h, char *key)
{
	void *r = NULL;
	struct hashnode *desired;

	assert(NULL != h);
	assert(NULL != key);

	desired = find_node(h, key);
	
	/* can set r to dummy tail value member (NULL) */
	if (desired)
		r = desired->value;

	return r;
}

void
insert_value(struct hashtable *h, char *key, void *value)
{
	unsigned long hv;
	unsigned long index;
	struct hashnode *head, *n = malloc(sizeof(*n));

	assert(NULL != h);
	assert(NULL != key);

	n->key = malloc(strlen(key)+1);
	strcpy(n->key, key);
	n->value = value;

	hv = hash((unsigned char *)n->key);
	index = MOD(hv, h->count);

	/* dummy head node */
	head = h->buckets[index];

	n->next = head->next;
	n->prev = head;

	head->next->prev = n;
	head->next = n;

	++h->size;
}

void *
remove_key(struct hashtable *h, char *key)
{
	void *r = NULL;
	struct hashnode *n;

	assert(NULL != h);
	assert(NULL != key);

	if (NULL != (n = find_node(h, key)))
	{
		r = n->value;
		n->value = NULL;
		free(n->key);
		n->key = NULL;
		n->next->prev = n->prev;
		n->prev->next = n->next;
		n->next = n->prev = NULL;
		free(n);
		--h->size;
	}

	return r;
}

void
free_hashtable(struct hashtable *h)
{
	unsigned int i;

	assert(h);

	for (i = 0; i < h->count; ++i)
	{
		struct hashnode *chain = h->buckets[i];
		while (chain)
		{
			struct hashnode *tmp = chain->next;
			free(chain->key);
			/* does not free the value member */
			free(chain);
			chain = tmp;
		}
	}
	free(h->buckets);
	h->buckets = NULL;
	free(h);
	h = NULL;
}

/* djb2 hash function */
unsigned long
hash(unsigned char *str)
{
	unsigned long hash = 5381;
	unsigned int c;

	while ((c = *str++))
		hash = (hash * 33) ^ c;
	/*	hash = ((hash << 5) + hash) ^ c; */

	return hash;
}
