/*
	Copyright (C) 2004, Bruce Ediger

    This file is part of srf.

    srf is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <srf.h>
#include <hashtable.h>
#include <value.h>
#include <function.h>

#ident "$Id: value.c,v 1.8 2004/11/02 04:39:39 bediger Exp $"

struct hashtable *value_table = NULL;

static struct value *free_list = NULL;

void
setup_value_table(struct hashtable *h)
{
	value_table = h;
}

struct value *
value_lookup(const char *id)
{
	struct value *v = NULL;

	if (!value_table)
		value_table = init_hashtable(1024, 10);

	v = data_lookup(value_table, id);

	return v;
}

/* returns any struct value * previously associated with the string id */
struct value *
value_add(const char *id, struct value *v)
{
	struct value *r = NULL;
	struct hashnode *n = NULL;
	unsigned int hv;

	if (!value_table)
		value_table = init_hashtable(1024, 10);

	if ((n = node_lookup(value_table, id, &hv)))
	{
		r = (struct value *)n->data;
		n->data = v;
	} else {
		const char *p = add_data(value_table, id, (void *)v);
		/* since node_lookup() failed, this should fail, too */
		assert(NULL == p);
	}
	return r;
}

struct value *
new_value(void)
{
	struct value *r = NULL;
	if (free_list)
	{
		r = free_list;
		assert (NULL != r);
		free_list = r->next;
	} else {
		r = malloc(sizeof(*r));
		assert (NULL != r);
	}

	r->type = UNSET;
	r->v.identifier = NULL;
	r->v.value = 0;
	r->v.function = NULL;
	r->assigned = 0;
	r->next = NULL;

	return r;
}

struct value_list *
new_value_list(void)
{
	struct value_list *vl = NULL;
	vl = malloc(sizeof(*vl));
	if (vl)
	{
		vl->head = vl->tail = NULL;
		vl->cnt = 0;
		vl->assigned = 0;
	}
	return vl;
}

void
free_value_list(struct value_list *vl)
{
	if (vl && !vl->assigned)
	{
		free_value(vl->head);
		vl->head = vl->tail = NULL;
		vl->cnt = 0;
		vl->assigned = 0;
		free(vl);
	}
}

void
free_value(struct value *v)
{
	while (v)
	{
		struct value *tmp = v->next;
		if (!v->assigned)
		{
			if (FUNCTION == v->type && v->v.function)
			{
				free_function(v->v.function);
				v->v.function = NULL;
			}
			v->next = free_list;
			free_list = v;
		}
		v = tmp;
	}
}

void
free_values(void)
{
	while (free_list)
	{
		struct value *tmp = free_list->next;

		/* don't have to free() free_list->v.identifier: atom */

		if (FUNCTION == free_list->type)
		{
			free_function(free_list->v.function);
			free_list->v.function = NULL;
		}

		free_list->next = NULL;

		free(free_list);

		free_list = tmp;
	}
}

void
print_value(struct value *v)
{
	if (v)
	{
		switch (v->type)
		{
		case UNSET:
			printf("Unset value");
			break;
		case INTEGER:
			printf("%d", v->v.value);
			break;
		case FUNCTION:
			print_function(v->v.function);
			break;
		case IDENTIFIER:
			printf("\"%s\"", v->v.identifier);
			break;
		case FILENAME:
			printf("loading \"%s\"", v->v.filename);
		}
		printf("\n");
	} else {
	}
}

struct function *
value_function(struct value *v)
{
	struct function *r = NULL;
	if (v)
	{
		switch (v->type)
		{
		case UNSET:
			printf("Evaluating unset value as function pointer\n");
			break;
		case INTEGER:
			printf("Evaluating integer value %d for function pointer\n", v->v.value);
			break;
		case FILENAME:
			printf("Evaluating filename \"%s\" as function pointer\n",
				v->v.filename);
			break;
		case FUNCTION:
			r = v->v.function;
			break;
		case IDENTIFIER:
			r = value_function(value_lookup(v->v.identifier));
			break;
		}
	} else {
		printf("Evaluating NULL value as function pointer\n");
		r = new_function();
		r->function = NULL;
	}
	return r;
}

int
evaluate(struct value *v)
{
	int r = 0;
	if (v)
	{
		switch (v->type)
		{
		case UNSET:
			printf("Evaluating unset value as integer\n");
			break;
		case FILENAME:
			printf("Evaluating file name \"%s\" as integer\n", v->v.filename);
			break;
		case INTEGER:
			r = v->v.value;
			break;
		case FUNCTION:
			printf("Evaluating function pointer as integer\n");
			break;
		case IDENTIFIER:
			r = evaluate(value_lookup(v->v.identifier));
			break;
		}
	} else {
		printf("Evaluating NULL value as integer\n");
	}
	return r;
}

struct value *
evaluate_function(struct function *f, struct value_list *args)
{
	struct value *r = NULL;

	if (f)
	{
		if (NOFUNC == f->type)
		{
			printf("Evaluating unset function\n");
			r = new_value();
		} else if (f->function) {
			r = f->function(f, args);
			free_value_list(args);
		} else {
			printf("Evaluating NULL function pointer\n");
			r = new_value();
		}
	} else {
		printf("Evaluating NULL function\n");
		r = new_value();
	}

	return r;
}

struct value_list *
copy_and_evaluate(struct value_list *args)
{
	struct value_list *r = NULL;

	if (args)
	{
		r = new_value_list();
		r->cnt = 0;
		r->head = NULL;
		r->tail = NULL;
		if (args->head)
		{
			struct value *p = args->head;
			while (p)
			{
				struct value *t = copy_value(p);
				if (r->tail)
				{
					r->tail->next = t;
					r->tail = t;
				} else {
					r->head = r->tail = t;
				}
				++r->cnt;
				p = p->next;
			}
		}
	}

	return r;
}

struct value *
copy_value(struct value *original)
{
	struct value *r = NULL;

	if (original)
	{
		struct value *tv;
		r = new_value();
		r->type = original->type;
		switch (original->type)
		{
		case UNSET:
			break;
		case INTEGER:
			r->v.value = original->v.value;
			break;
		case FILENAME:
			r->v.filename = original->v.filename;
			break;
		case FUNCTION:
			/* XXX - does this wrong */
			r->v.function = copy_function(original->v.function);
			break;
		case IDENTIFIER:
			/* XXX - always give back an evaluated value */
			r->type = INTEGER;
			tv = value_lookup(original->v.identifier);
			r->v.value = evaluate(tv);
			free_value(tv);
			break;
		}
		r->assigned = 0;
		r->next = NULL;
	}

	return r;
}
